import pickle
import os
import time
from traceback import format_exc
import listener
from functools import wraps
from errno import ENOENT, EAGAIN, EACCES
from fcntl import lockf, LOCK_EX, LOCK_NB
from io import StringIO
from subprocess import Popen, PIPE, STDOUT
from csv import DictReader
from pwd import getpwnam
import locale
from univention.ox.oxlogger import OXLogger


SPOOL_DIR_PREFIX = '/var/spool/univention-directory-listener'
OX_BIN_DIRECTORY = '/opt/open-xchange/sbin'

OX_MASTER_ADMIN = "oxadminmaster"
OX_MASTER_SECRET = "/etc/ox-secrets/master.secret"

CSV_DELIMITER = ','
CSV_QUOTECHAR = '"'


class ListenerTools_ExceptionBase(Exception):
	pass


class ListenerTools_Unconfigured(ListenerTools_ExceptionBase):
	pass


class ListenerTools_RuntimeError(ListenerTools_ExceptionBase):
	def __init__(self, msg, exit_code=None, *args, **kwargs):
		self.exit_code = exit_code
		self.stderr = kwargs.pop('stderr', '')
		super(ListenerTools_RuntimeError, self).__init__(msg)


class LDAPObject(dict):
	"""
	LDAPObject behaves like a normal dictionary containing the attribute values of
	a LDAP object. In addition it is possible to read the DN of the LDAP object from
	a member attribute.

	>>> new = LDAPObject(dn='uid=Administrator,cn=users,dc=example,dc=com', uid=['Administrator'])
	>>> new.dn
	'uid=Administrator,cn=users,dc=example,dc=com'
	>>> new['uid']
	['Administrator']
	"""
	def __init__(self, dn, *args, **kwargs):
		self.dn = dn
		dict.__init__(self, *args, **kwargs)


class QueueEntry(object):
	"""
	This class represents one queue entry and will be pickled to disk for offline storage.
	It contains the old and new LDAP object (as "LDAPObject"), the action and the timestamp
	of creation.
	"""
	def __init__(self, dn, new, old, action, timestamp):
		self.new = LDAPObject(dn, new)
		self.old = LDAPObject(dn, old)
		self.action = action
		self.timestamp = timestamp

	@property
	def dn(self):
		"""
		Returns DN of LDAP object (preferably the "new" DN).
		"""
		if self.new and self.new.dn:
			return self.new.dn
		elif self.old and self.old.dn:
			return self.old.dn
		return None


class ListenerQueue(object):
	"""
	Example listener code:

	@ListenerQueue(module_name=name)
	def handler(dn, new, old, action=None):
		if not os.path.exists('/etc/foo.secret'):
			return ListenerTools_RuntimeError('this service is unconfigured')
		if not foo.service_is_available:
			raise ListenerTools_RuntimeError('the service is down')
		foo.do_something(dn, new, old)
	"""
	def __init__(self, module_name=None, stop_on_error=True):
		"""
		Creates a new ListenerQueue and loads existing pickle files into the queue.

		module_name:
			name of the listener module. The spooling directory is derived
			from the module name: '/var/spool/univention-directory-listener/$(module_name)s/'
		stop_on_error:
			If stop_on_error is True, the listener module stops processing queue entries
			on the first error/exception. Otherwise the module will skip the erroneous
			queue entry and continue processing the other one. Errorneous queue entries
			will be reprocessed on the next call of process().
		"""
		global name
		self.stop_on_error = stop_on_error
		if module_name:
			self.module_name = module_name
		else:
			try:
				self.module_name = name
			except NameError:
				raise ListenerTools_Unconfigured('The module_name has not been specified and could not be determined automatically!')
		self.spool_dir = os.path.join(SPOOL_DIR_PREFIX, self.module_name)
		self.queue = []
		self.logger = OXLogger(self.module_name)

		self.reread_queue_from_disk()

	@property
	def queue_size(self):
		return len(self.queue)

	def __call__(self, func):
		def process_pending():
			ent = getpwnam("listener")
			os.seteuid(ent.pw_uid)

			try:
				return self.process(func)
			except ListenerTools_RuntimeError as ex:
				self.logger.warn('Runtime error: {}'.format(ex))
				return

		@wraps(func)
		def wrapper(dn, new, old, action=None):
			self.add(dn, new, old, action)
			return self.process(func)

		wrapper.process_pending = process_pending
		return wrapper

	def add(self, dn, new, old, action=None):
		"""
		Add new entry to queue.
		"""
		# use timestamp of current time as filename
		# this way files/timestamps may be simply sorted later on
		timestamp = '%.10f' % time.time()
		entry = QueueEntry(dn, new, old, action, timestamp)
		self._save_entry(entry)
		self.queue.append(timestamp)

	def process(self, func):
		"""
		Process all available objects in queue. The specified function "func" is
		called with four arguments:
		dn: DN of the LDAP object
		new: new attributes
		old: old attributes
		action: modification type (only if specified during add)
		"""
		result = None
		dn = None

		with Exclusive(os.path.join(self.spool_dir, "lock.tmp")):
			queue_unprocessed = True
			while queue_unprocessed:
				queue_unprocessed = False

				for timestamp in self.queue:
					self.logger.info('Loading transaction {}'.format(timestamp))
					try:
						entry = self._load_entry(timestamp)
					except IOError as ex:
						if ex.errno != ENOENT:
							self.logger.error('loading queue entry {} failed (errno {})'.format(timestamp, ex.errno))
							raise
						self.reread_queue_from_disk()
						queue_unprocessed = True
						break

					try:
						dn = entry.dn
						result = func(dn, entry.new, entry.old, entry.action)
						self._delete_entry(timestamp)
						self.logger.info('transaction {} processed: dn={}'.format(timestamp, dn))
					except ListenerTools_RuntimeError as ex:
						self.logger.process('transaction deferred: dn={} {}'.format(dn, ex))
						if self.stop_on_error:
							break
					except Exception as ex:
						self.logger.error('handler failed: dn={}\n{}\n{}'.format(dn, format_exc(), ex))
						raise

			# some entries might have been removed, so the queue is reread
			self.reread_queue_from_disk()

		return result

	def clear_queue(self):
		for fn in os.listdir(self.spool_dir):
			if fn.endswith('.tmp'):
				continue
			os.remove(fn)

	def reread_queue_from_disk(self):
		if not os.path.isdir(self.spool_dir):
			self.queue = []
			return
		self.queue = [fn for fn in os.listdir(self.spool_dir) if not fn.endswith('.tmp')]
		sorted(self.queue)

	def _save_entry(self, entry):
		"""
		Write entry as pickle file to spool directory.
		"""

		# if spool directory is missing then create it
		if not os.path.isdir(self.spool_dir):
			os.makedirs(self.spool_dir, 0700)

		# write pickle file
		fn = os.path.join(self.spool_dir, entry.timestamp)
		fn_tmp = fn + '.tmp'
		with open(fn_tmp, 'wb') as fd:
			os.chmod(fn_tmp, 0600)
			pickle.dump(entry, fd, 2)
			fd.flush()
			os.fsync(fd.fileno())
		os.rename(fn_tmp, fn)

	def _load_entry(self, timestamp):
		"""
		Load pickle object from disk
		"""
		fn = os.path.join(self.spool_dir, timestamp)
		with open(fn, 'rb') as fd:
			return pickle.load(fd)

	def _delete_entry(self, timestamp):
		fn = os.path.join(self.spool_dir, timestamp)
		os.remove(fn)


class AsRoot(object):
	"""
	Temporarily change effective UID to 'root', if it's not already 'root'.
	"""
	def __init__(self):
		self.old_uid = None

	def __enter__(self):
		self.old_uid = os.geteuid()
		if self.old_uid != 0:
			listener.setuid(0)

	def __exit__(self, exc_type, exc_value, traceback):
		if self.old_uid == os.geteuid():
			return
		listener.unsetuid()


class Exclusive(object):
	def __init__(self, filename):
		self.lock_name = filename
		self.lock_file = None

	def __enter__(self):
		try:
			lock_file = os.open(self.lock_name, os.O_RDWR | os.O_CREAT, 0600)
			lockf(lock_file, LOCK_EX | LOCK_NB)
			self.lock_file = lock_file
		except IOError as ex:
			if ex.errno not in (EAGAIN, EACCES):
				raise
			raise ListenerTools_RuntimeError("concurrent process")

	def __exit__(self, exc_type, exc_value, traceback):
		if self.lock_file:
			os.close(self.lock_file)
			self.lock_file = None


def get_credentials(filename):
	"""
	Try to read secret file using effective UID 'root'.
	"""
	with AsRoot():
		try:
			with open(filename, 'r') as sfile:
				secret = sfile.read()
		except IOError as ex:
			raise ListenerTools_RuntimeError(ex)
	return secret.rstrip()


def ox_list_context(user, password, logger=None):
	return _ox_csvreader('listcontext', user, password, logger=logger)


def ox_list_user(user, password, context, searchpattern=None, ignorecase=False, logger=None):
	return _ox_csvreader('listuser', user, password, context, searchpattern, ignorecase, logger=logger)


def ox_list_group(user, password, context, searchpattern=None, logger=None):
	return _ox_csvreader('listgroup', user, password, context, searchpattern, logger=logger)


def _ox_csvreader(list_type, user, password, context=None, searchpattern=None, ignorecase=False, logger=None):
	"""
	Execute listuser/group, read output and return csv.DictReader.
	"""
	if not logger:
		logger = OXLogger('ox_csvreader')
		logger.stdout = True
	command = [
		os.path.join(OX_BIN_DIRECTORY, list_type),
		'--adminuser=%s' % user,
		'--adminpass=%s' % password,
		'--csv',
	]
	if context:
		command += ['--contextid', '%s' % context]
	if searchpattern:
		command += ['--searchpattern', '%s' % searchpattern]
	if ignorecase:
		command += ['--ignorecase']

	logger.info('command {}'.format(' '.join(_debug_command(command))))
	process = Popen(command, stdout=PIPE, stderr=PIPE)
	out, err = process.communicate()
	if process.returncode:
		raise ListenerTools_RuntimeError(
			'"{} {}" returned  {}'.format(list_type, context, process.returncode),
			process.returncode,
			stderr=err)

	out = out.decode(locale.getpreferredencoding()) # io.StringIO requires a unicode string
	memfile = StringIO(out)  # csv.DictReader wants a file-like-object
	return DictReader(memfile, delimiter=CSV_DELIMITER, quotechar=CSV_QUOTECHAR)


def ox_call(command, expected_error=None, logger=None):
	"""
	Execute command and raise RuntimeError on failures.
	"""
	if not logger:
		logger = OXLogger('ox_call')
		logger.stdout = True

	command = [os.path.join(OX_BIN_DIRECTORY, command[0])] + command[1:]
	logger.info('command {}'.format(' '.join(_debug_command(command))))
	process = Popen(command, stdout=PIPE, stderr=STDOUT)
	stdout, _stderr = process.communicate()
	if process.returncode == 0:
		return
	if expected_error and expected_error in stdout:
		logger.process('ignoring expected error "{}": {}'.format(expected_error, stdout.split('\n')[0].rstrip(' :')))
		return
	logger.error('failed with {}: {}'.format(process.returncode, stdout))
	raise ListenerTools_RuntimeError('"{}" returned  {}: {}'.format(command, process.returncode, stdout), process.returncode)


def _debug_command(command):
	"""
	Dump command excluding passwords.
	"""
	return [
		c for c in command
		if not c.startswith("--adminpass=") and not c.startswith("--password=")
	]


def listener_is_deactivated(name):
	if listener.configRegistry.is_false('ox/listener/enabled', False):
		OXLogger(name).process('listener module deactivated by UCR variable "ox/listener/enabled".')
		return True
	return False


if __name__ == '__main__':
	import doctest
	doctest.testmod()
